"""
Tweakers.net comments.
"""

from .user import User


class Comment:
    """
    A comment on a gathering.tweakers.net topic
    """

    def __init__(self, **kwargs) -> None:
        for k, v in kwargs.items():
            if k == "username":
                setattr(self, "user", User(name=v))
            else:
                setattr(self, k, v)



"""
gathering.tweakers.net topics
"""

import time

from typing import Union, List, Generator

from requests_html import HTMLResponse

from .utils import id_from_url, get
from .comment import Comment
from . import parsers


class Topic:
    """
    A topic on gathering.tweakers.net
    """

    def __init__(self, url, **kwargs) -> None:
        self.__dict__.update(kwargs)
        self.url: str = url
        self.id: int = id_from_url(self.url)

    def comments(self, page: Union[int, str]) -> List[Comment]:
        """
        Get comments for a specific Topic page

        :param page: Page number (zero indexed) or 'last' for last page.
        :return: A list of Comment objects.
        """
        response: HTMLResponse = get(url=f"{self.url}/{page}")
        return [Comment(**d) for d in parsers.topic_comments(response.html)]

    def comment_stream(
        self, refresh: int = 15, last: int = 3
    ) -> Generator[Comment, None, None]:
        """
        Generator of new comments.

        :param refresh: Refresh timer in seconds (default 15).
        :param last: Number of already posted comments to return (default 15).
        :return: A generator of Comment objects.
        """

        # get the last posted comment id, required for getting new comments
        comments: List = self.comments("last")
        last_message_id: int = comments[-1].id

        # yield last n comments
        for comment in comments[-last:]:
            yield comment

        timer: int = refresh
        while True:
            epoch_time: int = int(time.time())  # required for getting new comments
            ajax_url: str = f"https://gathering.tweakers.net/ajax/list_new_messages/{self.id}/{last_message_id}?output=json\
                    &nocache={epoch_time}"
            comments = self.get_new_comments(ajax_url)

            if comments:
                last_message_id = comments[-1].id
            for comment in comments:
                yield comment

            # decrease timer each second
            while timer > 0:  # pragma: no cover
                time.sleep(1)
                timer -= 1
            else:  # pragma: no cover
                timer = refresh

    def get_new_comments(self, ajax_url: str) -> List[Comment]:  # pragma: no cover
        """
        Get new comments for a given ajax_url

        :param ajax_url: Ajax url to query for new comments.
        :return: A list of comment objects.
        """
        json = get(ajax_url).json()

        new_comments: List
        try:
            html: str = "".join(json["data"]["messages"])
            new_comments = [Comment(**d) for d in parsers.topic_comments(html)]
        except KeyError:  # no new messages
            new_comments = []
        return new_comments



"""
https://tweakers.net frontpage.
"""

from typing import List

from requests_html import HTMLResponse
from .utils import get
from .topic import Topic
from .comment import Comment
from . import parsers


url = "https://tweakers.net"


class Article:
    def __init__(self, url, **kwargs):
        self.url = url
        self.__dict__.update(kwargs)

    def comments(self):
        """Get comments for a frontpage article.

        :return: A list of Comment objects.
        """
        # TODO: ADD PAGINATION OF COMMENTS
        response: HTMLResponse = get(url=f"{self.url}")
        return [Comment(**d) for d in parsers.article_comments(response.html)]


def articles() -> List["Articles"]:
    response: HTMLResponse = get(url)
    articles: List = [Article(**d) for d in parsers.frontpage_articles(response.html)]
    return articles


def active_topics() -> List[Topic]:
    response: HTMLResponse = get(url=f"{url}/forum/list_activetopics")
    topics: List = [Topic(**d) for d in parsers.active_topics(response.html)]
    return topics



"""
gathering.tweakers.net (forums).
"""

from typing import List

from requests_html import HTMLResponse
from .utils import get
from .topic import Topic
from . import parsers


url = "https://gathering.tweakers.net"


def active_topics() -> List[Topic]:
    response: HTMLResponse = get(url=f"{url}/forum/list_activetopics")
    topics: List = [Topic(**d) for d in parsers.active_topics(response.html)]
    return topics


def search(query: str) -> List[Topic]:
    """Search for topics given a query.
    :param query: Search query.
    """
    response: HTMLResponse = get(url=f"{url}/forum/find?keyword={query}")
    topics: List = [Topic(**d) for d in parsers.search_topics(response.html)]
    return topics


def bookmarks() -> List[Topic]:
    """Get a list of bookmarks

    :return: Bookmarks
    """
    response: HTMLResponse = get(url=f"{url}/forum/list_bookmarks")
    topics: List = [Topic(**d) for d in parsers.bookmark_topics(response.html)]
    return topics



class RateLimitException(Exception):
    """HTTP 429: Too Many Requests. This is a rate limit problem."""


class InvalidCredentialsException(Exception):
    """Invalid username or password."""


class CaptchaRequiredException(Exception):
    """Captcha warning triggered, unable to login."""



"""
HTML parsers.
"""

from typing import Dict, Generator, Union

from requests_html import HTML
import dateparser


def get_comment_count(html: HTML) -> int:
    comment_count: int
    try:
        comment_count = int(html.find(".commentCount", first=True).text)
    except AttributeError:  # Also count the opening post
        comment_count = 1
    return comment_count


def active_topics(html: HTML) -> Generator[dict, None, None]:
    for tr in html.find(".listing tr")[1:]:
        topic: Dict = {
            "title": tr.find(".topic a", first=True).text,
            "url": tr.find(".topic a", first=True).attrs["href"],
            "poster": tr.find(".poster", first=True).text,
            "last_reply": dateparser.parse(
                tr.find(".time a", first=True).text, languages=["nl"]
            ),
            "comment_count": get_comment_count(tr),
        }
        yield topic


def search_topics(html: HTML) -> Generator[dict, None, None]:
    for tr in html.find(".forumlisting tr")[2::2]:
        topic: Dict = {
            "title": tr.find(".title a", first=True).text,
            "url": tr.find(".title a", first=True).attrs["href"],
            "last_reply": dateparser.parse(
                tr.find(".time a", first=True).text, languages=["nl"]
            ),
        }
        yield topic


def bookmark_topics(html: HTML) -> Generator[dict, None, None]:
    for tr in html.find(".listing tr.alt")[1:]:
        topic: Dict = {
            "title": tr.find(".topic a")[1].text,
            "url": tr.find(".topic a")[1].attrs["href"],
            "last_reply": dateparser.parse(
                tr.find(".time a", first=True).text, languages=["nl"]
            ),
        }
        yield topic


def get_rating(div):
    rating: int
    try:
        rating = int(div.find("span.ratingcount", first=True).text.replace("+", ""))
    except AttributeError:  # post too old for rating
        rating = 0
    return rating


def topic_comments(html: Union[HTML, str]) -> Generator[dict, None, None]:
    if isinstance(html, str):
        html = HTML(html=html)

    for div in html.find(".message"):
        message: Dict = {
            "id": int(div.attrs["data-message-id"]),
            "username": div.find("a.user", first=True).text,
            "date": dateparser.parse(
                div.find("div.date", first=True).text, languages=["nl"]
            ),
            "url": div.find(".date p a", first=True).attrs["href"],
            "rating": get_rating(div),
            "text": div.find(".messagecontent", first=True).text,
            "html": div.html,
        }
        yield message


def frontpage_articles(html: HTML) -> Generator[dict, None, None]:
    for tr in html.find("tr.headline.news"):
        topic: Dict = {
            "title": tr.find(".title a", first=True).text,
            "url": tr.find(".title a", first=True).attrs["href"],
            "comment_count": get_comment_count(tr),
            "publication_time": tr.find(".publicationTime", first=True).text,
        }
        yield topic


def article_comments(html: HTML) -> Generator[dict, None, None]:
    for div in html.find("div.reactieBody"):
        comment: Dict = {
            "username": _get_text(div, selector=".userLink"),
            "date": dateparser.parse(
                _get_text(div, selector="a.date"), languages=["nl"]
            ),
            "text": _get_text(div, selector=".reactieContent"),
            "score": int(_get_text(div, selector="a.scoreButton")),
        }
        yield comment


def _get_text(html: HTML, selector: str) -> Union[str, None]:
    try:
        return html.find(selector, first=True).text
    except AttributeError:
        return



"""
A tweakers.net user.
"""


class User:
    """A Tweakers.net user"""

    def __init__(self, **kwargs) -> None:
        required = ["id", "name"]
        assert any(
            x in required for x in kwargs
        ), f"Missing one of required keywords: {required}"

        for k, v in kwargs.items():
            if k == "id":
                k = "_id"
            setattr(self, k, v)



"""
Utilities.
"""
from requests import Response
from tenacity import (
    retry,
    stop_after_attempt,
    wait_exponential,
    retry_if_exception_type,
)
from requests_html import HTMLSession, HTMLResponse

from tweakers.exceptions import (
    RateLimitException,
    InvalidCredentialsException,
    CaptchaRequiredException,
)

session = HTMLSession()
session.headers.update({"X-Cookies-Accepted": "1"})  # Bypass the cookiewall


@retry(
    retry=retry_if_exception_type(),
    wait=wait_exponential(multiplier=1, min=5, max=10),
    stop=stop_after_attempt(3),
)
def get(url: str) -> HTMLResponse:
    response = session.get(url)
    if response.status_code == 429:
        print("429, Rate limited...")
        raise RateLimitException()
    if not 200 <= response.status_code < 300:
        raise Exception(f"Url {url} returned a {response.status_code}")
    return response


def id_from_url(url: str) -> int:
    """
    Parse the id from the URL

    :param url: url to get id from.
    :raises NotImplementedError: If getting id from this url is not supported.
    :return: integer id.
    """
    parts = url.split("/")
    if "aanbod" in parts:
        id = parts[parts.index("aanbod") + 1]
    elif "list_messages" in parts:
        id = parts[parts.index("list_messages") + 1]
    elif "pricewatch" in parts:
        id = parts[parts.index("pricewatch") + 1]
    else:
        raise NotImplementedError(
            "Getting the id for url ({url}) is not yet implemented"
        )
    return int(id)


def login(username: str, password: str) -> None:  # pragma: no cover
    url = "https://tweakers.net/my.tnet/login/"
    response = session.get(url)
    try:
        token = response.html.find("input[name=tweakers_login_form\[_token\]]")[
            0
        ].attrs["value"]
    except IndexError:  # Already logged in
        return

    data = {
        "tweakers_login_form[_token]": token,
        "tweakers_login_form[user]": username,
        "tweakers_login_form[password]": password,
    }
    login_response = session.post(url=url, data=data)

    _raise_for_invalid_credentials(login_response)
    _raise_for_captcha_error(login_response)


def _raise_for_invalid_credentials(login_response: Response) -> None:
    login_error_text = (
        "De combinatie van gebruikersnaam of e-mailadres en wachtwoord is onjuist."
    )
    if login_error_text in login_response.text:
        raise InvalidCredentialsException


def _raise_for_captcha_error(login_response: Response) -> None:
    captcha_error_text = (
        "Om te bewijzen dat je geen robot bent, moet een captcha worden ingevuld."
    )
    if captcha_error_text in login_response.text:
        raise CaptchaRequiredException



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['tweakers']

package_data = \
{'': ['*']}

install_requires = \
['dateparser>=0.7.0,<0.8.0',
 'requests-html>=0.10.0,<0.11.0',
 'tenacity>=6.2.0,<7.0.0']

setup_kwargs = {
    'name': 'tweakers',
    'version': '0.3.3',
    'description': 'A Python wrapper for tweakers.net',
    'long_description': '# Tweakers\n![Build](https://github.com/timotk/tweakers/workflows/Build/badge.svg)\n[![codecov](https://codecov.io/gh/timotk/tweakers/branch/master/graph/badge.svg)](https://codecov.io/gh/timotk/tweakers)\n[![PyPI](https://img.shields.io/pypi/v/tweakers.svg)](https://pypi.org/project/tweakers)\n![PyPI - Python Version](https://img.shields.io/pypi/pyversions/tweakers.svg)\n\nA Python wrapper for [tweakers.net](https://tweakers.net/)\n\n## Install\n```\npip install tweakers\n```\n\n## Usage\n```\nimport tweakers\n```\n\n### Gathering\nWith the `tweakers.gathering` module you can access the forum.\n#### Active topics\n```\nfor topic in tweakers.gathering.active_topics():\n    print(topic.title)\n```\n\n#### Search\n```\nfor topic in tweakers.gathering.search(\'tweakers.net\'):\n    print(topic.title)\n```\n\n### Topic\n#### Get comments for a specific topic\n```\ntopic = Topic("https://gathering.tweakers.net/forum/list_messages/1551828")\nfor comment in topic.comments(page=1):\n    print(comment.user.name, comment.text)\n```\n\n#### Generate new comments as they are added\n```\nfor comment in topic.comment_stream():\n    print(comment.user.name, comment.text)\n```\n\n### Login\n```\ntweakers.login("YOUR_USERNAME", "YOUR_PASSWORD")\n```\nNow you can access pages that are unavailable for logged out users:\n```\nfor topic in tweakers.gathering.bookmarks():\n    print(topic.name)\n```\n',
    'author': 'Timo',
    'author_email': None,
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/timotk/tweakers',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.6',
}


setup(**setup_kwargs)
